package com.halomobi.ssp.base.utils

import android.database.Cursor
import android.os.Environment
import android.text.TextUtils
import com.halomobi.ssp.base.utils.Utils.context
import java.io.BufferedReader
import java.io.ByteArrayOutputStream
import java.io.Closeable
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.io.IOException
import java.io.InputStream
import java.io.InputStreamReader
import java.io.OutputStream
import java.io.RandomAccessFile
import java.io.UnsupportedEncodingException
import java.nio.channels.FileChannel
import java.security.MessageDigest
import java.util.Properties
import java.util.UUID

object FileUtils {
    const val DOWNLOAD_DIR: String = "download"
    const val CACHE_DIR: String = "cache"
    const val ICON_DIR: String = "icon"

    val isSDCardAvailable: Boolean
        /**
         * 判断SD卡是否挂载
         */
        get() = Environment.MEDIA_MOUNTED == Environment.getExternalStorageState()

    @JvmStatic
    val downloadDir: String?
        /**
         * 获取下载目录
         */
        get() = getDir(
            isSDCardAvailable,
            DOWNLOAD_DIR
        )

    val cacheDir: String?
        /**
         * 获取缓存目录
         */
        get() = getDir(
            isSDCardAvailable,
            CACHE_DIR
        )

    @JvmStatic
    val iconDir: String?
        /**
         * 获取icon目录
         */
        get() = getDir(
            isSDCardAvailable,
            ICON_DIR
        )

    /**
     * 获取应用目录，当SD卡存在时，获取SD卡上的目录，当SD卡不存在时，获取应用的cache目录
     */
    fun getDir(isSDCardAvailable: Boolean, name: String?): String? {
        try {
            val sb = StringBuilder()
            if (isSDCardAvailable) {
                sb.append(externalStoragePath)
            } else {
                sb.append(cachePath)
            }
            sb.append(name)
            sb.append(File.separator)
            val path = sb.toString()
            if (createDirs(path)) {
                return path
            } else if (isSDCardAvailable) {
                getDir(false, name)
            }
        } catch (e: Exception) {
            e.printStackTrace()
        }
        return null
    }

    val externalStoragePath: String
        /**
         * 获取SD下的应用目录
         */
        get() {
            val sb = StringBuilder()
            sb.append(Environment.getExternalStorageDirectory().absolutePath)
            sb.append(File.separator)
            sb.append("Android")
            sb.append(File.separator)
            sb.append("data")
            sb.append(File.separator)
            sb.append(context?.packageName)
            sb.append(File.separator)

            //        LogUtils.d("SD下的应用目录   :   " + sb.toString());
            return sb.toString()
        }

    val cachePath: String
        /**
         * 获取应用的cache目录
         */
        get() {
            val sb = StringBuilder()
            sb.append(File.separator)
            sb.append("data")
            sb.append(File.separator)
            sb.append("data")
            sb.append(File.separator)
            sb.append(context?.packageName)
            sb.append(File.separator)
            sb.append("cache")
            sb.append(File.separator)
            return sb.toString()
        }

    /**
     * 创建文件夹
     */
    @JvmStatic
    fun createDirs(dirPath: String): Boolean {
        if (!TextUtils.isEmpty(dirPath)) {
            val file = File(dirPath)
            if (file.exists() && !file.isDirectory) {
                if (file.delete()) {
//                    LogUtils.d("文件删除成功");
                } else {
//                    LogUtils.d("文件删除失败");
                }
            }
            return if (!file.exists()) {
                if (file.mkdirs()) {
                    //                    LogUtils.d("创建" + dirPath + "   成功");
                    true
                } else {
                    //                    LogUtils.d("创建" + dirPath + "   失败");
                    false
                }
            } else {
                true
            }
        }
        return false
    }

    /**
     * 复制文件，可以选择是否删除源文件
     */
    fun copyFile(srcPath: String, destPath: String, deleteSrc: Boolean): Boolean {
        val srcFile = File(srcPath)
        val destFile = File(destPath)
        return copyFile(srcFile, destFile, deleteSrc)
    }

    /**
     * 复制文件，可以选择是否删除源文件
     */
    fun copyFile(srcFile: File, destFile: File?, deleteSrc: Boolean): Boolean {
        if (!srcFile.exists() || !srcFile.isFile) {
            return false
        }
        var `in`: InputStream? = null
        var out: OutputStream? = null
        try {
            `in` = FileInputStream(srcFile)
            out = FileOutputStream(destFile)
            val buffer = ByteArray(1024)
            var i = -1
            while ((`in`.read(buffer).also { i = it }) > 0) {
                out.write(buffer, 0, i)
                out.flush()
            }
            if (deleteSrc && srcFile.exists() && !srcFile.delete()) {
                return false
            }
        } catch (e: Exception) {
            e.printStackTrace()
            return false
        } finally {
            close(out, `in`)
        }
        return true
    }

    fun getString4Asset(fileName: String): String? {
        var tBufferedReader: BufferedReader? = null
        try {
            val `is` = context?.assets?.open(fileName)
            tBufferedReader = BufferedReader(InputStreamReader(`is`, "UTF-8"))
            val tStringBuffer = StringBuffer()
            var sTempOneLine: String? = null
            while ((tBufferedReader.readLine().also { sTempOneLine = it }) != null) {
                tStringBuffer.append(sTempOneLine)
            }
            return tStringBuffer.toString()
        } catch (e: IOException) {
            e.printStackTrace()
        } finally {
            close(tBufferedReader)
        }
        return null
    }

    /**
     * 从输入流获取数据
     *
     * @param inputStream
     * @return
     * @throws Exception
     */
    @JvmStatic
    fun getByteFromStream(inputStream: InputStream?): ByteArray {
        val outputStream = ByteArrayOutputStream()
        try {
            val buffer = ByteArray(1024)
            var len = -1
            while ((inputStream?.read(buffer).also {
                    if (it != null) {
                        len = it
                    }
                }) != -1) {
                outputStream.write(buffer, 0, len)
            }
        } catch (e: IOException) {
            e.printStackTrace()
        } finally {
            close(outputStream, inputStream)
        }
        return outputStream.toByteArray()
    }


    /**
     * 关闭流
     */
    @JvmStatic
    fun close(vararg io: Closeable?) {
        for (closeable in io) {
            if (closeable != null) {
                try {
                    closeable.close()
                } catch (e: IOException) {
                    e.printStackTrace()
                }
            }
        }
    }

    /**
     * 关闭流
     */
    @JvmStatic
    fun close(cursor: Cursor?) {
        if (cursor != null) {
            try {
                cursor.close()
            } catch (e: Exception) {
                e.printStackTrace()
            }
        }
    }

    /**
     * 判断文件是否可写
     */
    fun isWriteable(path: String): Boolean {
        try {
            if (TextUtils.isEmpty(path)) {
                return false
            }
            val f = File(path)
            return f.exists() && f.canWrite()
        } catch (e: Exception) {
            e.printStackTrace()
            return false
        }
    }

    /**
     * 修改文件的权限,例如"777"等
     */
    fun chmod(path: String, mode: String) {
        try {
            val command = "chmod $mode $path"
            val runtime = Runtime.getRuntime()
            runtime.exec(command)
        } catch (e: Exception) {
            e.printStackTrace()
        }
    }

    /**
     * 把数据写入文件
     *
     * @param is       数据流
     * @param path     文件路径
     * @param recreate 如果文件存在，是否需要删除重建
     * @return 是否写入成功
     */
    fun writeFile(`is`: InputStream?, path: String, recreate: Boolean): Boolean {
        var res = false
        val f = File(path)
        var fos: FileOutputStream? = null
        try {
            if (recreate && f.exists() && !f.delete()) {
                return res
            }
            if (!f.exists() && null != `is` && createDirs(f.parent)) {
                var count = -1
                val buffer = ByteArray(1024)
                fos = FileOutputStream(f)
                while ((`is`.read(buffer).also { count = it }) != -1) {
                    fos.write(buffer, 0, count)
                }
                res = true
            }
        } catch (e: Exception) {
            e.printStackTrace()
        } finally {
            close(fos, `is`)
        }
        return res
    }

    /**
     * 把字符串数据写入文件
     *
     * @param content 需要写入的字符串
     * @param path    文件路径名称
     * @param append  是否以添加的模式写入
     * @return 是否写入成功
     */
    fun writeFile(content: ByteArray?, path: String, append: Boolean): Boolean {
        var res = false
        val f = File(path)
        var raf: RandomAccessFile? = null
        try {
            if (f.exists()) {
                if (!append && (!f.delete() || !f.createNewFile())) {
                    return res
                }
            } else if (!f.createNewFile()) {
                return res
            }
            if (f.canWrite()) {
                raf = RandomAccessFile(f, "rw")
                raf.seek(raf.length())
                raf.write(content)
                res = true
            }
        } catch (e: Exception) {
            e.printStackTrace()
        } finally {
            close(raf)
        }
        return res
    }

    /**
     * 把字符串数据写入文件
     *
     * @param content 需要写入的字符串
     * @param path    文件路径名称
     * @param append  是否以添加的模式写入
     * @return 是否写入成功
     */
    fun writeFile(content: String, path: String, append: Boolean): Boolean {
        try {
            return writeFile(content.toByteArray(charset("UTF-8")), path, append)
        } catch (e: UnsupportedEncodingException) {
            e.printStackTrace()
        }
        return false
    }

    /**
     * 把键值对写入文件
     *
     * @param filePath 文件路径
     * @param key      键
     * @param value    值
     * @param comment  该键值对的注释
     */
    fun writeProperties(filePath: String, key: String?, value: String?, comment: String?) {
        if (TextUtils.isEmpty(key) || TextUtils.isEmpty(filePath)) {
            return
        }
        var fis: FileInputStream? = null
        var fos: FileOutputStream? = null
        val f = File(filePath)
        try {
            if ((!f.exists() || !f.isFile) && !f.createNewFile()) {
                return
            }
            fis = FileInputStream(f)
            val p = Properties()
            p.load(fis) // 先读取文件，再把键值对追加到后面
            p.setProperty(key, value)
            fos = FileOutputStream(f)
            p.store(fos, comment)
        } catch (e: Exception) {
            e.printStackTrace()
        } finally {
            close(fis, fos)
        }
    }

    /**
     * 根据值读取
     */
    fun readProperties(filePath: String, key: String?, defaultValue: String?): String? {
        if (TextUtils.isEmpty(key) || TextUtils.isEmpty(filePath)) {
            return null
        }
        var value: String? = null
        var fis: FileInputStream? = null
        val f = File(filePath)
        try {
            if ((!f.exists() || !f.isFile) && !f.createNewFile()) {
                return null
            }
            fis = FileInputStream(f)
            val p = Properties()
            p.load(fis)
            value = p.getProperty(key, defaultValue)
        } catch (e: IOException) {
            e.printStackTrace()
        } finally {
            close(fis)
        }
        return value
    }

    /**
     * 把字符串键值对的map写入文件
     */
    fun writeMap(filePath: String, map: Map<String?, String?>?, append: Boolean, comment: String?) {
        if (map == null || map.size == 0 || TextUtils.isEmpty(filePath)) {
            return
        }
        var fis: FileInputStream? = null
        var fos: FileOutputStream? = null
        val f = File(filePath)
        try {
            if ((!f.exists() || !f.isFile) && !f.createNewFile()) {
                return
            }
            val p = Properties()
            if (append) {
                fis = FileInputStream(f)
                p.load(fis) // 先读取文件，再把键值对追加到后面
            }
            p.putAll(map)
            fos = FileOutputStream(f)
            p.store(fos, comment)
        } catch (e: Exception) {
            e.printStackTrace()
        } finally {
            close(fis, fos)
        }
    }

    /**
     * 把字符串键值对的文件读入map
     */
    fun readMap(filePath: String, defaultValue: String?): Map<String, String>? {
        if (TextUtils.isEmpty(filePath)) {
            return null
        }
        var map: Map<String, String>? = null
        var fis: FileInputStream? = null
        val f = File(filePath)
        try {
            if ((!f.exists() || !f.isFile) && !f.createNewFile()) {
                return map
            }
            fis = FileInputStream(f)
            val p = Properties()
            p.load(fis)
            map =p as HashMap<String,String> // 因为properties继承了map，所以直接通过p来构造一个map
        } catch (e: Exception) {
            e.printStackTrace()
        } finally {
            close(fis)
        }
        return map
    }

    /**
     * 改名
     */
    fun copy(src: String, des: String, delete: Boolean): Boolean {
        val file = File(src)
        if (!file.exists()) {
            return false
        }
        val desFile = File(des)
        var `in`: FileInputStream? = null
        var out: FileOutputStream? = null
        try {
            `in` = FileInputStream(file)
            out = FileOutputStream(desFile)
            val buffer = ByteArray(1024)
            var count = -1
            while ((`in`.read(buffer).also { count = it }) != -1) {
                out.write(buffer, 0, count)
                out.flush()
            }
        } catch (e: Exception) {
            e.printStackTrace()
            return false
        } finally {
            close(`in`, out)
        }
        if (delete) {
            if (file.delete()) {
//                LogUtils.d("文件删除成功");
            } else {
//                LogUtils.d("文件删除失败");
            }
        }
        return true
    }

    /**
     * 复制文件，支持断点续传
     *
     * @param in
     * @param destFile
     * @return
     */
    fun copyFile(`in`: InputStream?, destFile: File): Boolean {
        if (`in` == null) {
            return false
        }
        var out: RandomAccessFile? = null
        try {
            out = RandomAccessFile(destFile, "rw")
            out.seek(destFile.length())
            val buffer = ByteArray(1024)
            var i = -1
            while ((`in`.read(buffer).also { i = it }) > 0) {
                out.write(buffer, 0, i)
            }
        } finally {
            close(out, `in`)
        }
        return true
    }

    fun getFileMD5String(file: File): String {
        var `in`: FileInputStream? = null
        try {
            `in` = FileInputStream(file)
            val ch = `in`.channel
            val byteBuffer = ch.map(
                FileChannel.MapMode.READ_ONLY, 0,
                file.length()
            )
            val messagedigest = MessageDigest.getInstance("MD5")
            messagedigest.update(byteBuffer)
            val md5str = KeyUtil.bytesToHexString(messagedigest.digest())
            return md5str
        } catch (e: Exception) {
            e.printStackTrace()
        } finally {
            close(`in`)
        }
        return UUID.randomUUID().toString()
    }
}
