package com.halomobi.ssp.base.utils

import android.annotation.SuppressLint
import android.content.Context
import android.net.ConnectivityManager
import android.net.wifi.WifiManager
import android.os.Build
import android.telephony.TelephonyManager
import android.text.TextUtils
import com.halomobi.ssp.base.utils.FileUtils.close
import com.halomobi.ssp.base.utils.SystemUtils.getInfoByJson
import com.halomobi.ssp.base.utils.Utils.context
import com.halomobi.ssp.base.utils.Utils.sp
import com.halomobi.ssp.model.request.iptask.RequestIpManager
import java.io.BufferedReader
import java.io.InputStreamReader
import java.net.NetworkInterface
import java.util.Collections

object NetworkUtils {
    const val NET_NONE: Int = -1
    const val NET_UNKNOWN: Int = 0x00
    const val NET_WIFI: Int = 0x02
    const val NET_CELL_UNKNOWN: Int = 0x03
    const val NET_2G: Int = 0x04
    const val NET_3G: Int = 0x05
    const val NET_4G: Int = 0x06

    val netType: Int
        /**
         * 判断当前是否有网
         *
         *
         * 1 Wifi
         *
         *
         * 2 4G
         *
         *
         * 3 3G
         *
         *
         * 4 2G
         *
         *
         * 0 无网
         *
         * @return 判断值
         */
        get() {
            val cwjManager =
                context?.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
            val info = cwjManager.activeNetworkInfo
                ?: return 0 // 无网
            val netType = info.type
            if (netType == ConnectivityManager.TYPE_WIFI) {
                return 2 // wifi
            } else if (netType == ConnectivityManager.TYPE_MOBILE) {
                return when (info.subtype) {
                    TelephonyManager.NETWORK_TYPE_GPRS, TelephonyManager.NETWORK_TYPE_EDGE, TelephonyManager.NETWORK_TYPE_CDMA, TelephonyManager.NETWORK_TYPE_1xRTT, TelephonyManager.NETWORK_TYPE_IDEN -> 3
                    TelephonyManager.NETWORK_TYPE_UMTS, TelephonyManager.NETWORK_TYPE_EVDO_0, TelephonyManager.NETWORK_TYPE_EVDO_A, TelephonyManager.NETWORK_TYPE_HSDPA, TelephonyManager.NETWORK_TYPE_HSUPA, TelephonyManager.NETWORK_TYPE_HSPA, TelephonyManager.NETWORK_TYPE_EVDO_B, TelephonyManager.NETWORK_TYPE_EHRPD, TelephonyManager.NETWORK_TYPE_HSPAP -> 4
                    TelephonyManager.NETWORK_TYPE_LTE -> 5
                    else -> 0
                }
            }
            return 0
        }

    /**
     * 当前网络是否可用
     *
     * @param cxt Context
     * @return activeNetInfo.isConnected();
     */
    @JvmStatic
    fun isNetworkConnected(cxt: Context?): Boolean {
        val connectivityManager = cxt
            ?.applicationContext?.getSystemService(
                Context.CONNECTIVITY_SERVICE
            ) as ConnectivityManager

        val activeNetInfo = connectivityManager.activeNetworkInfo
        if (activeNetInfo != null) {
            return activeNetInfo.isConnected
        }
        return false
    }

    @JvmStatic
    val networkType: Int
        /**
         * 判断当前是否有网
         *
         *
         * 2 Wifi
         *
         *
         * 6 4G
         *
         *
         * 5 3G
         *
         *
         * 4 2G
         *
         *
         * -1 无网
         *
         * @return 判断值
         */
        get() {
            val cwjManager =
                context?.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
            if (cwjManager != null) {
                val info = cwjManager.activeNetworkInfo
                    ?: return NET_NONE // 无网
                val netType = info.type
                if (netType == ConnectivityManager.TYPE_WIFI) {
                    return NET_WIFI // wifi
                } else if (netType == ConnectivityManager.TYPE_MOBILE) {
                    return getMobileNetClass(info.subtype) // 各种手机网络
                }
            }
            return 0
        }

    /**
     * 区分手机制式
     *
     * @param networkType
     * @return
     */
    private fun getMobileNetClass(networkType: Int): Int {
        return when (networkType) {
            TelephonyManager.NETWORK_TYPE_GPRS, TelephonyManager.NETWORK_TYPE_EDGE, TelephonyManager.NETWORK_TYPE_CDMA, TelephonyManager.NETWORK_TYPE_1xRTT, TelephonyManager.NETWORK_TYPE_IDEN -> NET_2G
            TelephonyManager.NETWORK_TYPE_UMTS, TelephonyManager.NETWORK_TYPE_EVDO_0, TelephonyManager.NETWORK_TYPE_EVDO_A, TelephonyManager.NETWORK_TYPE_HSDPA, TelephonyManager.NETWORK_TYPE_HSUPA, TelephonyManager.NETWORK_TYPE_HSPA, TelephonyManager.NETWORK_TYPE_EVDO_B, TelephonyManager.NETWORK_TYPE_EHRPD, TelephonyManager.NETWORK_TYPE_HSPAP -> NET_3G
            TelephonyManager.NETWORK_TYPE_LTE -> NET_4G
            else -> NET_UNKNOWN
        }
    }


    val macAddress: String
        /**
         * 获取当前设备的MAC地址
         */
        get() {
            val mac = getInfoByJson("mac")
            return mac
            //        if (!TextUtils.isEmpty(mac)) return mac;
//        if (SystemUtils.getSDK_INT() >= 23) {
//            return getWifiMacAddress();
//        } else {
//            return getMacInNormal();
//        }
        }

    val wifiMacAddress: String
        /**
         * 获取wifi mac地址
         *
         * @return wifi mac地址
         */
        get() {
            try {
                val interfaceName = "wlan0"
                val interfaces: List<NetworkInterface> =
                    Collections.list(NetworkInterface.getNetworkInterfaces())
                for (intf in interfaces) {
                    if (!intf.name.equals(interfaceName, ignoreCase = true)) {
                        continue
                    }
                    var mac: ByteArray? = ByteArray(0)
                    mac = intf.hardwareAddress
                    if (mac == null) {
                        return ""
                    }
                    val buf = StringBuilder()
                    for (aMac in mac) {
                        buf.append(String.format("%02X:", aMac))
                    }
                    if (buf.length > 0) {
                        buf.deleteCharAt(buf.length - 1)
                    }
                    return buf.toString()
                }
            } catch (ex: Exception) {
            } // for now eat exceptions

            return ""
        }

    private val macInNormal: String
        @SuppressLint("MissingPermission", "HardwareIds")
        get() {
            var mac = ""
            val context = context
            if (null != context) {
                try {
                    val wm = context.applicationContext
                        .getSystemService(Context.WIFI_SERVICE) as WifiManager
                    val info = wm.connectionInfo
                    mac = info?.macAddress.toString()
                } catch (e: Exception) {
                    e.printStackTrace()
                }
            }
            return mac
        }


    private val macInAndroidM: String
        get() {
            var cmd: ProcessBuilder? = null
            var process: Process? = null
            var reader: BufferedReader? = null
            try {
                val args =
                    arrayOf("/system/bin/cat", "/sys/class/net/wlan0/address")
                cmd = ProcessBuilder(*args)
                process = cmd.start()
                reader = BufferedReader(
                    InputStreamReader(
                        process.inputStream,
                        "UTF-8"
                    )
                )
                val line = reader.readLine()
                return line
            } catch (ex: Exception) {
                ex.printStackTrace()
            } finally {
                close(reader)
                process?.destroy()
            }
            return ""
        }

    @JvmStatic
    val ipAddress: String?
        /**
         * 获得设备ip地址
         */
        get() {
            val ipAddress = getInfoByJson("ip")
            if (!TextUtils.isEmpty(ipAddress)) {
                return ipAddress
            }
            return sp?.getString(RequestIpManager.IP_ADDRESS, "")
            //        try {
//            Enumeration<NetworkInterface> en = NetworkInterface
//                    .getNetworkInterfaces();
//            while (en.hasMoreElements()) {
//                NetworkInterface intf = en.nextElement();
//                Enumeration<InetAddress> enumIpAddr = intf.getInetAddresses();
//                while (enumIpAddr.hasMoreElements()) {
//                    InetAddress inetAddress = enumIpAddr.nextElement();
//                    if (!inetAddress.isLoopbackAddress()) {
//                        return inetAddress.getHostAddress();
//                    }
//                }
//            }
//        } catch (SocketException e) {
//            e.printStackTrace();
//        }
//        return "";
        }

    @JvmStatic
    val netWorkID: String
        /**
         * 获取Wifi ID
         *
         * @return
         */
        get() {
            try {
                val wifi_service =
                    context?.applicationContext
                        ?.getSystemService(Context.WIFI_SERVICE) as WifiManager
                val wifiInfo = wifi_service.connectionInfo
                return wifiInfo.networkId.toString() + ""
            } catch (e: Exception) {
                return "nul"
            }
        }

    @JvmStatic
    val wifiMac: String
        /**
         * 获取局域网名称
         *
         * @return
         */
        @SuppressLint("MissingPermission")
        get() {
            try {
                val wifi_service =
                    context?.applicationContext
                        ?.getSystemService(Context.WIFI_SERVICE) as WifiManager
                val wifiInfo = wifi_service.connectionInfo
                return wifiInfo?.macAddress.toString()
            } catch (e: Exception) {
                return "nul"
            }
        }

    @JvmStatic
    val sSID: String
        /**
         * 获取局域网名称
         *
         * @return
         */
        get() {
            try {
                val wifi_service =
                    context?.applicationContext
                        ?.getSystemService(Context.WIFI_SERVICE) as WifiManager
                val wifiInfo = wifi_service.connectionInfo
                return wifiInfo.ssid
            } catch (e: Exception) {
                return "nul"
            }
        }

    @JvmStatic
    val lksd: Int
        /**
         * 连接速度
         *
         * @return
         */
        get() {
            try {
                val wifi_service =
                    context?.applicationContext
                        ?.getSystemService(Context.WIFI_SERVICE) as WifiManager
                val wifiInfo = wifi_service.connectionInfo
                return wifiInfo.linkSpeed
            } catch (e: Exception) {
                return 0
            }
        }

    @JvmStatic
    val rssi: Int
        /**
         * 接收信号强度
         *
         * @return
         */
        get() {
            try {
                val wifi_service =
                    context?.applicationContext
                        ?.getSystemService(Context.WIFI_SERVICE) as WifiManager
                val wifiInfo = wifi_service.connectionInfo
                return wifiInfo.rssi
            } catch (e: Exception) {
                return 0
            }
        }

    @JvmStatic
    val isRoaming: Int
        /**
         * 是否漫游
         *
         * @return
         */
        get() {
            try {
                val manager =
                    context?.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
                val info = manager.activeNetworkInfo
                if (info == null || !info.isConnected) {
                    return 0
                }
                return if (info.isRoaming) {
                    // here is the roaming option you can change it if you want to
                    // disable internet while roaming, just return false
                    1
                } else {
                    0
                }
            } catch (e: Exception) {
                e.printStackTrace()
                return 0
            }
        }
}
