package com.halomobi.ssp.bussiness.module_sdk.view.imageloader

import android.annotation.SuppressLint
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.drawable.BitmapDrawable
import android.os.Build
import android.text.TextUtils
import android.view.View
import android.widget.ImageView
import com.halomobi.ssp.base.utils.FileUtils.getByteFromStream
import com.halomobi.ssp.base.utils.KeyUtil.bytesToHexString
import com.halomobi.ssp.base.utils.Utils.runInMainThread
import com.halomobi.ssp.bussiness.module_sdk.view.gif.GifView
import com.halomobi.ssp.bussiness.module_sdk.view.imageloader.cache.DoubleCache
import com.halomobi.ssp.bussiness.module_sdk.view.imageloader.cache.ImageCache
import com.halomobi.ssp.framework.schedule.ITask.LaunchMode
import com.halomobi.ssp.framework.schedule.ThreadManager
import com.halomobi.ssp.framework.schedule.ThreadManager.ThreadPoolProxy
import com.halomobi.ssp.model.net.absnet.AbsNetTask
import com.halomobi.ssp.model.net.http.HttpResult
import java.io.InputStream
import java.net.HttpURLConnection

object ImageLoader {
    /**
     * 图片下载的线程池名称
     */
    private const val THREAD_POOL_NAME = "IMAGE_THREAD_POOL"
    private val mOptions = BitmapFactory.Options()

    /**
     * 图片下载的线程池
     */
    private val mThreadPool: ThreadPoolProxy = ThreadManager.getSinglePool(THREAD_POOL_NAME)
    private val imageCache: ImageCache = DoubleCache()

    //	private static volatile ImageLoader instance;DownloadDBManager
    init {
        mOptions.inDither = false // 设置为false，将不考虑图片的抖动值，这会减少图片的内存占用
        mOptions.inPurgeable = true // 设置为true，表示允许系统在内存不足时，删除bitmap的数组。
        mOptions.inInputShareable =
            true // 和inPurgeable配合使用，如果inPurgeable是false，那么该参数将被忽略，表示是否对bitmap的数组进行共享
    }

    /**
     * 加载图片
     */
    @JvmStatic
    fun load(view: View?, url: String?) {
        if (view == null || TextUtils.isEmpty(url)) {
            return
        }
        view.tag = url // 把控件和图片的url进行绑定，因为加载是一个耗时的，等加载完毕了需要判定该控件是否和该url匹配
        val data = imageCache.getCache(url)
        if (data != null) {
            setImageSafe(view, url, data) // 如果内存中加载到了，直接设置图片
        } else {
            asyncLoad(view, url)
        }
    }

    /**
     * 异步加载图片
     */
    private fun asyncLoad(view: View, url: String?) {
        // 先创建一个runnable封装执行过程
        val runnable = Runnable {
            val data = loadFromNet(url) // 如果内存中加载到了，直接设置图片
            if (data.isNotEmpty()) {
                setImageSafe(view, url, data)
            }
        }
        mThreadPool.execute(runnable) // 执行任务
    }

    fun getBitmap(url: String): Bitmap? {
        var bitmap: Bitmap? = null
        val array = getByteArray(url)
        if (array != null && array.size > 0) {
            bitmap = BitmapFactory.decodeByteArray(array, 0, array.size)
        }
        return bitmap
    }

    @JvmStatic
    fun getByteArray(url: String): ByteArray? {
        var data: ByteArray? = ByteArray(0)
        try {
            if (!TextUtils.isEmpty(url)) {
                data = imageCache.getCache(url)
                if (data == null || data.isEmpty()) {
                    data = loadFromNet(url) // 如果内存中加载到了，直接设置图片
                }
                return data
            }
        } catch (e: OutOfMemoryError) {
            imageCache.clear()
            e.printStackTrace()
        } catch (e: Exception) {
            e.printStackTrace()
        }
        return data
    }

    /**
     * 从网络加载图片
     */
    private fun loadFromNet(url: String?): ByteArray {
        var data = ByteArray(0)
        var httpResult: HttpResult? = null
        try {
            httpResult = HttpResult(object : AbsNetTask(ReqType.Get) {

                override val entity: ByteArray
                    get() = ByteArray(0)

                override val url: String?
                    get() = url

                override fun onAddHeaders(urlConnection: HttpURLConnection?) {

                }

                override val launchMode: LaunchMode
                    get() = LaunchMode.ADD_NEW


                override val name: String
                    get() = "ImageLoader"


            }.openConnection(url))
        } catch (e: Exception) {
            e.printStackTrace()
        }
        var stream: InputStream? = null
        if (httpResult != null && (httpResult.inputStream.also { stream = it }) != null) { // 请求网络
            data = getByteFromStream(stream)
            httpResult.close()
            if (data.isNotEmpty()) {
                imageCache.putCache(url, data)
            }
        }
        return data
    }

    @JvmStatic
    fun adapterData(view: View?, data: ByteArray?): Bitmap? {
        var bitmap: Bitmap? = null
        val dst = ByteArray(4)
        data?.let { System.arraycopy(it, 0, dst, 0, 4) }
        val fileType = bytesToHexString(dst)
        if (!fileType.equals(
                "47494638",
                ignoreCase = true
            ) || view?.javaClass != GifView::class.java
        ) { // 非gif数据或者非gif渲染控件
            bitmap = data?.size?.let { BitmapFactory.decodeByteArray(data, 0, it, mOptions) }
        }
        return bitmap
    }

    /**
     * 设置给控件图片
     */
    @SuppressLint("NewApi")
    private fun setImageSafe(
        view: View?, url: String?,
        data: ByteArray?
    ) {
        if (view == null || view.tag == null || TextUtils.isEmpty(url)
            || data == null || data.size == 0
        ) {
            return
        }

        val bitmap = adapterData(view, data)
        // 检测文件头
        runInMainThread(
            Runnable
            // 需要在主线程中设置
            {
                val tag: Any // 在主线程中判断，可以做到同步
                if ((view.tag.also { tag = it }) != null && tag is String) {
                    if (TextUtils.equals(tag, url)) { // 检测如果url和控件匹配
                        val sdk = Build.VERSION.SDK_INT
                        if (bitmap != null) {
                            if (view is ImageView) {
                                view.setImageBitmap(bitmap)
                                return@Runnable
                            }

                            if (sdk < Build.VERSION_CODES.JELLY_BEAN) {
                                view.setBackgroundDrawable(BitmapDrawable(bitmap))
                            } else {
                                view.background = BitmapDrawable(bitmap) // 就进行图片设置
                            }
                        } else if (view is GifView) {
                            val gifView = view
                            gifView.setShowDimension(gifView.measuredWidth, gifView.measuredHeight)
                            gifView.setGifImage(data)
                        }
                    }
                }
            })
    }
}
