package com.halomobi.ssp.framework.schedule

import java.util.concurrent.Executors
import java.util.concurrent.LinkedBlockingQueue
import java.util.concurrent.ThreadPoolExecutor
import java.util.concurrent.TimeUnit

object ThreadManager {
    private const val DEFAULT_SINGLE_POOL_NAME = "DEFAULT_SINGLE_POOL_NAME"

    private var mLongPool: ThreadPoolProxy? = null

    private var mShortPool: ThreadPoolProxy? = null

    private var mDownloadPool: ThreadPoolProxy? = null

    private var mCheckViewPool: ThreadPoolProxy? = null

    private var mRequestAdPool: ThreadPoolProxy? = null

    private val mMap: MutableMap<String, ThreadPoolProxy> = HashMap()
    private val mSingleLock = Any()

    @get:Synchronized
    val requestAdPool: ThreadPoolProxy
        /**
         * 请求广告线程
         */
        get() {
            if (mRequestAdPool == null) {
                mRequestAdPool = ThreadPoolProxy(3, 5, 5L)
            }
            return mRequestAdPool as ThreadPoolProxy
        }

    @get:Synchronized
    val downloadPool: ThreadPoolProxy
        /**
         * 获取下载线程
         */
        get() {
            if (mDownloadPool == null) {
                mDownloadPool = ThreadPoolProxy(3, 3, 5L)
            }
            return mDownloadPool as ThreadPoolProxy
        }

    @get:Synchronized
    val checkViewPool: ThreadPoolProxy
        get() {
            if (mCheckViewPool == null) {
                mCheckViewPool = ThreadPoolProxy(3, 3, 5L)
            }
            return mCheckViewPool as ThreadPoolProxy
        }

    @get:Synchronized
    val longPool: ThreadPoolProxy
        /**
         * 获取一个用于执行长耗时任务的线程池，避免和短耗时任务处在同一个队列而阻塞了重要的短耗时任务，通常用来联网操作
         */
        get() {
            if (mLongPool == null) {
                mLongPool = ThreadPoolProxy(3, 5, 5L)
            }
            return mLongPool as ThreadPoolProxy
        }

    @JvmStatic
    @get:Synchronized
    val shortPool: ThreadPoolProxy
        /**
         * 获取一个用于执行短耗时任务的线程池，避免因为和耗时长的任务处在同一个队列而长时间得不到执行，通常用来执行本地的IO/SQL
         */
        get() {
            if (mShortPool == null) {
                mShortPool = ThreadPoolProxy(2, 2, 5L)
            }
            return mShortPool as ThreadPoolProxy
        }

    val singlePool: ThreadPoolProxy
        /**
         * 获取一个单线程池，所有任务将会被按照加入的顺序执行，免除了同步开销的问题
         */
        get() = getSinglePool(DEFAULT_SINGLE_POOL_NAME)

    /**
     * 获取一个单线程池，所有任务将会被按照加入的顺序执行，免除了同步开销的问题
     */
    fun getSinglePool(name: String): ThreadPoolProxy {
        synchronized(mSingleLock) {
            var singlePool = mMap[name]
            if (singlePool == null) {
                singlePool = ThreadPoolProxy(1, 1, 1L)
                mMap[name] = singlePool
            }
            return singlePool
        }
    }

    fun onDestroy() {
        val entries: Set<Map.Entry<String, ThreadPoolProxy>> = mMap.entries
        for ((_, value) in entries) {
            value.shutdown()
        }
        mMap.clear()
        checkViewPool.stop()
    }

    class ThreadPoolProxy(
        private val mCorePoolSize: Int,
        private val mMaximumPoolSize: Int,
        private val mKeepAliveTime: Long
    ) {
        private var mPool: ThreadPoolExecutor? = null

        /**
         * 执行任务，当线程池处于关闭，将会重新创建新的线程池
         */
        @Synchronized
        fun execute(task: Runnable?) {
            if (task == null) {
                return
            }
            if (mPool == null || mPool?.isShutdown == true) {
                mPool = ThreadPoolExecutor(
                    mCorePoolSize, mMaximumPoolSize, mKeepAliveTime,
                    TimeUnit.SECONDS, LinkedBlockingQueue(),
                    Executors.defaultThreadFactory(), ThreadPoolExecutor.AbortPolicy()
                )
            }
            mPool?.execute(task)
        }

        /**
         * 取消线程池中某个还未执行的任务
         */
        @Synchronized
        fun cancel(run: Runnable?): Boolean {
            return mPool != null && (!((mPool?.isShutdown)?:false) || mPool?.isTerminating == true) && mPool?.queue?.remove(
                run
            ) == true
        }

        /**
         * 判断队列中是否包含该任务
         */
        @Synchronized
        fun contains(run: Runnable?): Boolean {
            return mPool != null && (!(mPool?.isShutdown?:false) || mPool?.isTerminating == true) && mPool?.queue?.contains(
                run
            ) == true
        }

        /**
         * 立刻关闭线程池，并且正在执行的任务也将会被中断
         */
        @Synchronized
        fun stop() {
            if (mPool != null && (mPool?.isShutdown != true || mPool?.isTerminating == true)) {
                mPool?.shutdownNow()
            }
        }

        /**
         * 平缓关闭单任务线程池，但是会确保所有已经加入的任务都将会被执行完毕才关闭
         */
        @Synchronized
        fun shutdown() {
            if (mPool != null && (mPool?.isShutdown != true || mPool?.isTerminating == true)) {
                mPool?.shutdown()
            }
        }
    }
}
