package com.halomobi.ssp.model.net.absnet

import android.text.TextUtils
import com.halomobi.ssp.base.common.ConstantPool
import com.halomobi.ssp.base.common.ConstantPool.EroType
import com.halomobi.ssp.base.utils.FileUtils.close
import com.halomobi.ssp.base.utils.LogUtils.d
import com.halomobi.ssp.base.utils.LogUtils.getThrowable
import com.halomobi.ssp.base.utils.NetworkUtils.isNetworkConnected
import com.halomobi.ssp.base.utils.Utils.context
import com.halomobi.ssp.base.utils.Utils.sp
import com.halomobi.ssp.framework.schedule.ITask
import com.halomobi.ssp.model.net.http.HttpResult
import com.halomobi.ssp.model.net.http.HttpResult.Companion.getMsgByErrorCode
import com.halomobi.ssp.model.net.http.MyX509TrustManager
import com.halomobi.ssp.model.net.http.TrustAnyHostnameVerifier
import com.halomobi.ssp.model.request.bean.DnsBean
import java.io.FileNotFoundException
import java.io.IOException
import java.io.OutputStream
import java.net.HttpURLConnection
import java.net.SocketTimeoutException
import java.net.URL
import java.net.UnknownHostException
import java.security.KeyManagementException
import java.security.NoSuchAlgorithmException
import javax.net.ssl.HttpsURLConnection
import javax.net.ssl.SSLContext
import javax.net.ssl.TrustManager

/**
 * 网络加载抽象任务类
 */
abstract class AbsNetTask(private val reqType: ReqType) : ITask {
    private var mListener: NetTaskListener? = null

    var mDnsBean: DnsBean? = null

    private var isGzipRequest = false

    /**
     * 是否通过IP对应关系重发网络请求
     */
    private var isReTry = false

    /**
     * 设置是否可以通过IP对应关系进行重发网络请求
     *
     * @param reTry 是否重发
     */
    protected fun setReTry(reTry: Boolean) {
        isReTry = reTry
    }

    override fun run() {
        if (!isNetworkConnected(context)) {
            d("netTask no network")
            if (mListener != null) {
                mListener?.onNetworkError()
            }
            return
        }
        execute()
    }

    protected abstract val entity: ByteArray?

    protected abstract val url: String?

    fun setListener(netTaskListener: NetTaskListener?) {
        this.mListener = netTaskListener
    }

    protected abstract fun onAddHeaders(urlConnection: HttpURLConnection?)

    private fun execute() {
        var httpResult: HttpResult? = null
        try {
            val url = url
            val connection = openConnection(url)
            httpResult = getHttpResult(connection, 1, 1)
            if (httpResult == null) {
                return
            }
            val resultCode = httpResult.code
            if (resultCode?.compareTo(200)==-1 || resultCode?.compareTo(400)==1 ||resultCode?.compareTo(400)==0) { // 大于400 网络错误
                onErrorReceived("NetException:" + getMsgByErrorCode(resultCode), resultCode)
                return
            }
            if (mListener != null) { //200<=resultCode<=300代表请求成功
                mListener?.onDataReceived(httpResult)
            }
        } catch (e: UnknownHostException) {
            onErrorReceived(getThrowable(e), EroType.ERROR_UNKNOWN_HOST)
            e.printStackTrace()
        } catch (e: FileNotFoundException) {
            onErrorReceived(getThrowable(e), EroType.ERROR_FILE_NOT_FOUND)
            e.printStackTrace()
        } catch (e: Exception) {
            e.printStackTrace()
            onErrorReceived(getThrowable(e), EroType.EXCEPTION)
        } finally {
            httpResult?.close()
        }
    }

    private fun getHttpResult(
        connection: HttpURLConnection?,
        RedirectsNum: Int,
        reTryNum: Int
    ): HttpResult? {
        var redirectNum = RedirectsNum
        var reTNum = reTryNum
        if (connection == null) {
            if (mListener != null) {
                mListener?.onNetworkError()
            }
            return null
        }

        if (RedirectsNum >= 5) {//循环重定向超过5次
            if (mListener != null) {
                mListener?.onErrorReceived("循环重定向超过5次： " + url, EroType.EXCEPTION)
            }
            return null
        }

        var httpResult: HttpResult?
        try {
            httpResult = HttpResult(connection)
            val resultCode = httpResult.code
            // 广告请求成功
            if (resultCode in 301..399) {
                val location = connection.getHeaderField("Location")
                if (!TextUtils.isEmpty(location)) {
                    httpResult = getHttpResult(openConnection(location), ++redirectNum, reTryNum)
                }
            }
        } catch (e: SocketTimeoutException) {
            if (reTryNum >= 3) {
                throw e
            }
            httpResult = getHttpResult(openConnection(url), RedirectsNum, ++reTNum)
            e.printStackTrace()
        }
        return httpResult
    }

    protected fun setGzipRequest(isGzipRequest: Boolean) {
        this.isGzipRequest = isGzipRequest
    }

    fun openConnection(url: String?): HttpURLConnection? {
        if (TextUtils.isEmpty(url)) {
            return null
        }

        val mUrl = URL(url)
        val mHttpStr = url?.substring(0, url.indexOf(":"))
        val connection = if (mHttpStr == "https") {
            getHttpsURLConnection(mUrl)
        } else {
            getHttpURLConnection(mUrl)
        }
        return dealWithHttpConnection(connection)
    }


    private fun dealWithHttpConnection(connection: HttpURLConnection): HttpURLConnection {
        if (isGzipRequest) {
            connection.addRequestProperty("Content-Type", "text/html")
            connection.addRequestProperty("Accept-Encoding", "gzip,deflate")
        } else {
            connection.addRequestProperty("Accept-Encoding", "")
            connection.setRequestProperty("Content-Type", "application/json;charset=utf-8")
        }
        val ua = sp?.getString(ConstantPool.UA_KEY, "")
        if (!TextUtils.isEmpty(ua)) {
            connection.addRequestProperty("User-agent", ua)
        }
        onAddHeaders(connection)
        when (reqType) {
            ReqType.Get -> connection.requestMethod = "GET"
            ReqType.Post -> {
                connection.requestMethod = "POST"
                connection.doOutput = true
                connection.doInput = true
                var outputStream: OutputStream? = null
                try {
                    outputStream = connection.outputStream
                    val entity = entity
                    if (entity != null && entity.size > 0) {
                        outputStream.write(entity)
                    }
                } catch (e: Exception) {
                    e.printStackTrace()
                } finally {
                    close(outputStream)
                }
            }

        }
        return connection
    }

    private fun getHttpURLConnection(url: URL): HttpURLConnection {
        val httpURLConnection = url.openConnection() as HttpURLConnection
        httpURLConnection.connectTimeout = SOCKET_TIMEOUT
        httpURLConnection.readTimeout = HTTP_TIMEOUT
        httpURLConnection.instanceFollowRedirects = true
        httpURLConnection.setRequestProperty("Accept-Charset", "UTF-8")
        return httpURLConnection
    }


    private fun getHttpsURLConnection(url: URL): HttpsURLConnection {
        val httpsURLConnection = url.openConnection() as HttpsURLConnection
        httpsURLConnection.connectTimeout = SOCKET_TIMEOUT
        httpsURLConnection.readTimeout = HTTP_TIMEOUT
        val tm = arrayOf<TrustManager>(MyX509TrustManager())
        val sslContext = SSLContext.getInstance("SSL")
        sslContext.init(null, tm, null)
        System.setProperty("https.protocols", "TLSv1,TLSv1.1,TLSv1.2")
        httpsURLConnection.sslSocketFactory = sslContext.socketFactory
        httpsURLConnection.hostnameVerifier = TrustAnyHostnameVerifier()
        httpsURLConnection.instanceFollowRedirects = true
        httpsURLConnection.setRequestProperty("Accept-Charset", "UTF-8")
        return httpsURLConnection
    }

    enum class ReqType {
        Get, Post
    }

    /**
     * 请求发生错误
     *
     * @param eroMsg  错误信息
     * @param eroCode 错误码
     */
    private fun onErrorReceived(eroMsg: String, eroCode: Int) {
        if (mListener != null) {
            mListener?.onErrorReceived(eroMsg, eroCode)
        }
    }

    companion object {
        /**
         * HTTP链接超时时间
         */
        private const val HTTP_TIMEOUT = 100 * 1000

        /**
         * 请求获取数据超时时间
         */
        private const val SOCKET_TIMEOUT = 100 * 1000
    }
}

